import hashlib
import logging
import struct

import sqlalchemy as sa


log = logging.getLogger(__name__)

__all__ = [
    'Locker',
    'AcquireFailure',
]


class AcquireFailure(Exception):
    pass


class Locker:
    """
        A Locker instance is intended to be an app-level lock factory.

        It holds the name of the application (so lock names are namespaced and less likely to
        collide) and the SQLAlchemy engine instance (and therefore the connection pool).
    """
    def __init__(self, app_name, db_url=None, blocking_default=True, acquire_timeout_default=30000,
            create_engine_callable=None):
        self.app_name = app_name
        self.blocking_default = blocking_default
        self.acquire_timeout_default = acquire_timeout_default

        if create_engine_callable:
            self.engine = create_engine_callable()
        else:
            self.engine = sa.create_engine(db_url)

        @sa.event.listens_for(self.engine, 'checkin')
        def on_conn_checkin(dbapi_connection, connection_record):
            """
                This function will be called when a connection is checked back into the connection
                pool.  That should happen when .close() is called on it or when the connection
                proxy goes out of scope and is garbage collected.
            """
            if dbapi_connection is None:
                # This may occur in rare circumstances where the connection is already closed or an
                # error occurred while connecting to the database. In these cases any held locks
                # should already be released when the connection terminated.
                return

            with dbapi_connection.cursor() as cur:
                # If the connection is "closed" we want all locks to be cleaned up since this
                # connection is going to be recycled.  This step is to take extra care that we don't
                # accidentally leave a lock acquired.
                cur.execute('select pg_advisory_unlock_all()')

    def _lock_name(self, name):
        if self.app_name is None:
            return name

        return '{}.{}'.format(self.app_name, name)

    def _lock_num(self, name):
        """
            PostgreSQL requires lock ids to be integers.  It accepts bigints which gives us
            64 bits to work with.  Hash the lock name to an integer.
        """
        name = self._lock_name(name)

        name_hash = hashlib.sha1(name.encode('utf-8'))

        # Convert the hash to an integer value in the range of a PostgreSQL bigint
        num, = struct.unpack('q', name_hash.digest()[:8])

        return num

    def lock(self, name, **kwargs):
        lock_num = self._lock_num(name)
        kwargs.setdefault('blocking', self.blocking_default)
        kwargs.setdefault('acquire_timeout', self.acquire_timeout_default)
        return Lock(self.engine.connect(), lock_num, **kwargs)


class Lock:
    def __init__(self, engine, lock_num, blocking=None, acquire_timeout=None):
        self.engine = engine
        self.conn = None
        self.lock_num = lock_num
        self.blocking = blocking
        self.acquire_timeout = acquire_timeout

    def acquire(self, blocking=None, acquire_timeout=None):
        blocking = blocking if blocking is not None else self.blocking
        acquire_timeout = acquire_timeout or self.acquire_timeout

        if self.conn is None:
            self.conn = self.engine.connect()

        if blocking:
            timeout_sql = sa.text('set lock_timeout = :timeout')
            self.conn.execute(timeout_sql, timeout=acquire_timeout)

            lock_sql = sa.text('select pg_advisory_lock(:lock_num)')
        else:
            lock_sql = sa.text('select pg_try_advisory_lock(:lock_num)')

        try:
            result = self.conn.execute(lock_sql, lock_num=self.lock_num)
            retval = result.scalar()
            log.debug('Lock result was: %r', retval)
            # At least on PG 10.6, pg_advisory_lock() returns an empty string
            # when it acquires the lock.  pg_try_advisory_lock() returns True.
            # If pg_try_advisory_lock() fails, it returns False.
            return retval in (True, '')
        except sa.exc.OperationalError as e:
            if 'lock timeout' not in str(e):
                raise
            log.debug('Lock acquire failed due to timeout')
            return False

    def release(self):
        if self.conn is None:
            return False

        sql = sa.text('select pg_advisory_unlock(:lock_num)')
        result = self.conn.execute(sql, lock_num=self.lock_num)
        self.conn.close()
        self.conn = None
        return result.scalar()

    def __enter__(self):
        if not self.acquire():
            raise AcquireFailure
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()

    def __del__(self):
        # Do everything we can to release resources and the connection to avoid accidentally holding
        # a lock indefinitely if .release() is forgotten.
        try:
            self.release()
        except Exception:
            # Sometimes this will fail if the connection has gone away before the gc runs.  Since
            # Python is just going to print the exception and we can't do anything about it,
            # suppress the exception to keep erroneous noise out of stderr.
            pass
        try:
            self.conn.close()
        except Exception:
            # ditto
            pass
        del self.conn



import datetime as dt
import gc
import os
import random
import string

import pytest

import pals

# Default URL will work for CI tests
db_url = os.environ.get('PALS_DB_URL', 'postgresql://postgres:password@localhost/postgres')


def random_str(length):
    return ''.join(random.choice(string.printable) for _ in range(25))


class TestLocker:

    def test_lock_num_generation(self):
        """
            Create 5000 random strings of varying lengths, convert them to their lock number, and
            make sure we still have 5000 unique numbers.  Given that any application using locking
            is probably going to have, at most, locks numbering in the tens or very low hundreds,
            this seems like an ok way to test that the method we are using to convert our strings
            into numbers is unlikely to have accidental collisions.
        """
        locker = pals.Locker('TestLocker', db_url)

        names = [random_str(max(6, x % 25)) for x in range(5000)]
        assert len(set(names)) == 5000
        nums = [locker._lock_num(name) for name in names]
        assert len(set(nums)) == 5000

    def test_locker_defaults(self):
        lock = pals.Locker('foo', db_url).lock('a')
        assert lock.blocking is True
        assert lock.acquire_timeout == 30000

        lock = pals.Locker('bar', db_url, blocking_default=False, acquire_timeout_default=1000) \
            .lock('a')
        assert lock.blocking is False
        assert lock.acquire_timeout == 1000


def duration(started_at):
    duration = dt.datetime.now() - started_at
    secs = duration.total_seconds()
    # in milliseconds
    return secs * 1000


class TestLock:
    locker = pals.Locker('TestLock', db_url, acquire_timeout_default=1000)

    def test_same_lock_fails_acquire(self):
        lock1 = self.locker.lock('test_it')
        lock2 = self.locker.lock('test_it')

        try:
            assert lock1.acquire() is True

            # Non blocking version should fail immediately.  Use `is` test to make sure we get the
            # correct return value.
            assert lock2.acquire(blocking=False) is False

            # Blocking version should fail after a short time
            start = dt.datetime.now()
            acquired = lock2.acquire(acquire_timeout=300)
            waited_ms = duration(start)

            assert acquired is False
            assert waited_ms >= 300 and waited_ms < 350
        finally:
            lock1.release()

    def test_different_lock_name_both_acquire(self):
        lock1 = self.locker.lock('test_it')
        lock2 = self.locker.lock('test_it2')

        try:
            assert lock1.acquire() is True
            assert lock2.acquire() is True
        finally:
            lock1.release()
            lock2.release()

    def test_lock_after_release_acquires(self):
        lock1 = self.locker.lock('test_it')
        lock2 = self.locker.lock('test_it')

        try:
            assert lock1.acquire() is True
            assert lock1.release() is True
            assert lock2.acquire() is True
        finally:
            lock1.release()
            lock2.release()

    def test_class_params_used(self):
        """
            If blocking & timeout params are set on the class, make sure they are passed through and
            used correctly.
        """
        lock1 = self.locker.lock('test_it')
        lock2 = self.locker.lock('test_it', blocking=False)
        lock3 = self.locker.lock('test_it', acquire_timeout=300)

        try:
            assert lock1.acquire() is True

            # Make sure the blocking param applies
            acquired = lock2.acquire()
            assert acquired is False

            # Make sure the retry params apply
            start = dt.datetime.now()
            acquired = lock3.acquire()
            waited_ms = duration(start)
            assert acquired is False
            assert waited_ms >= 300 and waited_ms < 350
        finally:
            lock1.release()
            lock2.release()
            lock3.release()

    def test_context_manager(self):
        lock2 = self.locker.lock('test_it', blocking=False)
        try:
            with self.locker.lock('test_it'):
                assert lock2.acquire() is False

            # Outside the lock should have been released and we can get it now
            assert lock2.acquire()
        finally:
            lock2.release()

    def test_context_manager_failure_to_acquire(self):
        """
            By default, we want a context manager's failure to acquire to be a hard error so that
            a developer doesn't have to remember to explicilty check the return value of acquire
            when using a with statement.
        """
        lock2 = self.locker.lock('test_it', blocking=False)
        assert lock2.acquire() is True

        with pytest.raises(pals.AcquireFailure):
            with self.locker.lock('test_it'):
                pass  # we should never hit this line

    def test_gc_lock_release(self):
        lock1 = self.locker.lock('test_it')
        lock1.acquire()
        del lock1
        gc.collect()

        assert self.locker.lock('test_it', blocking=False).acquire()



"""
    If you run this file and monitor connections on your PostgreSQL server, you should see a backend
    process dedicated to this lock:

        select
            stat.datname
            , stat.pid
            , stat.usename
            , stat.application_name
            , stat.state
            , locks.mode as lockmode
            , stat.query
        from pg_stat_activity stat
        left join pg_locks locks
            on locks.pid = stat.pid
        where locktype = 'advisory'

    Then kill the Python process running this file in any way you want and make sure the PostgreSQL
    pid with the lock goes away:

    1) press any key and exit
    2) CTRL-C
    3) kill -9 <pid>
"""
import os

import pals

locker = pals.Locker('pals-hang', 'postgresql://postgres:password@localhost:54321/postgres')

lock = locker.lock('hang')
lock.acquire()
print('My pid is: ', os.getpid())
input('Lock acquired, press any key to exit: ')
